/*************************************************
*
*  This program has been written by: Luigi Carnelli
*  Dipartimento di Scienze della Informazione
*  Universita' degli Studi di Milano
*  Laboratorio di Eidomatica
*
*  For any info or comment please contact:
*  prof. Daniele Marini
*  e_mail: marini@eidomatica.dsi.unimi.it
*
**************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <ctype.h>
#include <gl/gl.h>

#include "frcmpr.h"



static short red[256], blue[256], green[256];

#define SAVE_CMAP()	for (i=0; i<256; i++) getmcolor( i, &red[i], &blue[i], &green[i]);
#define RESTORE_CMAP()	for (i=0; i<256; i++) mapcolor(i, red[i], blue[i], green[i]); gflush();
#define CLEARS()	color(0);  clear(); 

#define OPENGRAPH(title);   prefsize(img_size_x*zoom, img_size_y*zoom); foreground();\
			    winopen( title ); ortho2(-.5, img_size_x*zoom-.5, img_size_y*zoom-.5, -.5);\
			    SAVE_CMAP();\
			    for (i=0; i<248; i++)\
				mapcolor(i,   i, i, i); mapcolor(BIANCO, 255, 255, 255);\
				mapcolor(NERO, 0, 0, 0); mapcolor(MAGENTA, 255, 0, 255);\
				mapcolor(ROSSO, 255, 0, 0); mapcolor(GIALLO, 255, 255, 0);\
				mapcolor(VERDE, 0, 255, 0); mapcolor(BLU, 0, 0, 255);\
				mapcolor(CIANO, 0, 255, 255); gflush(); CLEARS();

#define CLOSEGRAPH();   beep(); getchar(); getchar(); RESTORE_CMAP(); gexit();


int    rd_comp( const void *, const void * );


/* global's */
PAR_COMPR      par_cmp;
LIST           *lst = NULL;
LIST_X         *lst_x = NULL;
RAN_DOM        rda[5000];
CONTORNO       *cont;
BYTE           *image, *D_scaled, *image2, msg_buf[256];
int            img_size_x, img_size_y, D_size_x, D_size_y, 
               maxbox=1024, nbox=0, fzoom=1, zoom=3,  xxx;

static int     g_b_S[4], g_b2_S[4];
static int     mult_tab[1280];

#define        TARGET_ERROR    0
#define        TARGET_COMPR    1
static int     compr_mode = TARGET_ERROR;



/************/
void main()
/************/
{
FILE           *fp;
double         err;
int            i, j, k, x, y;
char           s[64], sr[64], mp[64], omap[64], c;
static char    *str = "\n\
   change compression (P)arameters \n\
   (V)isualise  \" \"\n\
   (L)oad        \" \"\n\
   show map (E)lements\n\
        \" \" (S)hape\n\
   \n\
   (C)ompress\n\
   (D)decompress\n\
   (-) show image difference \n\
   \n\
   e(X)it\n\n\n";


/* Compression parameters: default values*/
par_cmp.max_rms        = 8;			/*max value of RMS error */
par_cmp.D_R_rapp       = 2.0;			/* domain/range ratio  */
par_cmp.max_contrast   = 1.0;
par_cmp.min_div_area   = 16;			/* min size to subdivide an range*/
par_cmp.soglia         = 80;		    /* threshold to select edges*/
par_cmp.search_step    = 8;
par_cmp.complex        = 64;


                   start:
puts( str );
scanf( " %c", &c);
getchar();

switch( toupper(c) ) {
   /************* */
   case SET_PAR:
   /************* */
       printf( "Compr. mode: 0=target err, 1=target compr. ", compr_mode ); fflush( stdout );
	  /* target err select the quality of compression as the target, while
	  terget compr selects the compression ratio as the target */
       gets( s ); if( isdigit( *s ) )  sscanf( s, " %d", &compr_mode );
       printf( "Max RMS acceptable [%.1f] : ", par_cmp.max_rms ); fflush( stdout );
       gets( s ); if( isdigit( *s ) )  sscanf( s, " %f", &par_cmp.max_rms );
       printf( "Ratio between domain and range [%.1f] : ", par_cmp.D_R_rapp ); fflush( stdout );
       gets( s ); if( isdigit( *s ) )  sscanf( s, " %f", &par_cmp.D_R_rapp );
       printf( "Max contrast (abs) [%.1f] : ", par_cmp.max_contrast ); fflush( stdout );
       gets( s ); if( isdigit( *s ) )  sscanf( s, " %f", &par_cmp.max_contrast );
       printf( "Minimum size of range area [%d] : ", par_cmp.min_div_area ); fflush( stdout );
       gets( s ); if( isdigit( *s ) )  sscanf( s, " %d", &par_cmp.min_div_area );
       printf( "Threshold to extract edges [%d] : ", par_cmp.soglia ); fflush( stdout );
       gets( s ); if( isdigit( *s ) )  sscanf( s, " %d", &par_cmp.soglia );
       printf( "Box complexity [%d] : ", par_cmp.complex ); fflush( stdout );
       gets( s ); if( isdigit( *s ) )  sscanf( s, " %d", &par_cmp.complex );
       printf( "Step of domain search [%d] : ", par_cmp.search_step ); fflush( stdout );
       gets( s ); if( isdigit( *s ) )  sscanf( s, " %d", &par_cmp.search_step );

       crea_tab();
       goto start;

   /************* */
   case SHOW_PAR:
   /************* */
       printf( "Compr. mode: target %s\n", compr_mode?"compression":"error" ); fflush( stdout );
       printf( "Max value of RMS [%.1f]\n", par_cmp.max_rms ); fflush( stdout );
       printf( "Ratio between domain and range [%.1f]\n", par_cmp.D_R_rapp ); fflush( stdout );
       printf( "Max contrast [%.1f]\n", par_cmp.max_contrast ); fflush( stdout );
       printf( "Minimum size of range area [%d]\n", par_cmp.min_div_area ); fflush( stdout );
       printf( "Threshold to extract edges [%d]\n", par_cmp.soglia ); fflush( stdout );
       printf( "Box complexity [%d]\n", par_cmp.complex ); fflush( stdout );
       printf( "Step of domain search [%d]\n", par_cmp.search_step ); fflush( stdout );

       goto start;

   /************* */
   case LOAD_PAR:
   /************* */
       if( get_map_name( omap ) ) {
           fp = fopen( omap, "rb" );
           load_default_par();
           }
       goto start;

   /************* */
   case COMPRIMI_GREY:
   /************* */
       ask_str( "Image Name [no ext]: ", sr );
       strcat( sr, ".sgi" ); validate( sr );
       ask_str( "Map Name [no ext]: ", omap );
       strcat( omap, ".map" );
       ask_int( "Zoom [3]: ", zoom );
       
       load_image( sr );
       OPENGRAPH( "Grey Compression" );
       show_image();
       
       if( compr_mode == TARGET_ERROR ) 	/* if the compression target is 
									ruled by the quality of the 
									compression .... */
           comprimi_img( omap );
       else {							/* otherwise choose the number of
										maps, therefore fixing the
										compression ratio */
           ask_int( "Number of maps [1024]: ", maxbox );
           comprimi_img_2( omap );
           }

       free( image ); free( image2 ); CLOSEGRAPH();
       goto start;

   /************* */
   case DECOMPRIMI_GREY:
   /************* */
       ask_str( "Map name [no ext]: ", omap );
       strcat( omap, ".map" ); validate( omap );
       ask_int( "Zoom by pixel replication [3]: ", zoom );
       ask_int( "Fractal zoom [3]: ", fzoom );
       get_dim_fm_map( omap );

       img_size_x *= fzoom,  img_size_y *= fzoom;
       image = malloc( img_size_x * img_size_y );
       OPENGRAPH( "Grey Decompression" );

       for( i=0;   i < img_size_y;  i++ )
           for( j=0;   j < img_size_x;  j++ )
               IMG( j, i ) = rand() % 256;
       show_image();

       for( i=0;  i < 5;  i++ ) {          /* # of iterations... */
           show_compr_data( omap, i );
	   printf("Pass %d of %d\r", i+1, 5); fflush(stdout);
           }
       save_image( "rigen.sgi" );		/* the decompressed image is saved
								with the fixed name rigen.sgi */
       free( image ); CLOSEGRAPH();
       goto start;

   /************* */
   case SHOW_GREY_DIFF:
   /************* */
       ask_str( "Name of original image [no ext]: ", sr );
       strcat( sr, ".sgi" ); validate( sr );
       ask_str( "Name of decompressed image [no ext]: ", s );
       strcat( s, ".sgi" ); validate( s );
       ask_int( "Zoom [3]: ", zoom );

       err = diff_image( sr, s );

       OPENGRAPH( "Grey Diff" );
       show_image();
       MSG1VAL( "\nMSE: %.2lf", err );
       free( image ); free( image2 ); CLOSEGRAPH();
       goto start;

   /************* */
   case SHOW_MAP:
   /************* */
       if( ! get_img_name( s ) )
           goto start;
       if( ! get_map_name( omap ) )
           goto start;
       ask_int( "Zoom [3]: ", zoom );

       load_image( s );
       OPENGRAPH( "Show Grey Map Elements" );
       show_image();
       show_map( omap );
       free( image );  CLOSEGRAPH();
       goto start;

   /************* */
   case SHOW_SHAPE:
   /************* */
       if( ! get_map_name( omap ) )
           goto start;
       ask_int( "Zoom [3]: ", zoom );

       show_map_shape( BIANCO, omap );
       goto start;

   /************* */
   case QUIT:
   /************* */
	break;
	
   default:
       goto start;
   }
}




/************/
show_compr_data( char *mapname, int cnt )
/************/
{
RECT_DATA  *c;
RECT       r, d;
FILE       *fp;
double     o, s;
int        mode, i, j, x, y, v, rx, ry, cdim_x, cdim_y;

fp = fopen( mapname, "rb" );
load_default_par();
img_size_x *= fzoom, img_size_y *= fzoom;
crea_tab();
if( cnt == 0 ) show_cmp_par();

while( TRUE ) {
   readmap();

   r.x1*=fzoom,  r.y1*=fzoom,  rx*=fzoom,  ry*=fzoom,  d.x1*=fzoom,  d.y1*=fzoom;
   r.x2 = r.x1 + rx - 1,  r.y2 = r.y1 + ry - 1;
   c = alloc_rect( &r ); cdim_x = X_DIM(c);  cdim_y = Y_DIM(c);

   /* Set scale ratio to the range dimension ... */
   for( i=0;   i < cdim_y;   i++ ) {
       y = d.y1 + mult_tab[i];
       for( j=0;   j < cdim_x;   j++ ) {
           x = d.x1 + mult_tab[j];
           C(j, i) = (BYTE )( ((int )IMG(x, y)  +  (int )IMG(x+1, y) + 
                               (int )IMG(x, y+1) + (int )IMG(x+1, y+1) ) >> 2);
           }
       }

   for( i=0;   i < cdim_y;   i++ ) {
       y = i+r.y1;
       for( j=0;   j < cdim_x;   j++ ) {
           x = j+r.x1;
           switch( mode ) {			/* selects the type of rotation 
								transformation among four possible
								simple rotations */
               case 0:                 /* nessuna trasformazione... */
                   v = blimit( C( j, i ) * s + o ); break;
               case 1:                 /* inverti le X... */
                   v = blimit( C( cdim_x-1 - j, i ) * s + o ); break;
               case 2:                 /* inverti le Y... */
                   v = blimit( C( j, cdim_y-1 - i ) * s + o ); break;
               case 3:                 /* inverti le X e le Y... */
                   v = blimit( C( cdim_x-1 - j, cdim_y-1 - i ) * s + o ); break;
               }

           if( i == 0  &&  y-1 >= 0 ) {
               IMG( x, y ) = (BYTE)( .6*v + .4*IMG( x, y-1 ));
               fast_show_rect( blimit((int )(IMG(x,y))),  x,  y );
               }
           else if( j == 0  &&  x-1 >= 0 ) {
               IMG( x, y ) = (BYTE)( .6*v + .4*IMG( x-1, y ));
               fast_show_rect( blimit((int )(IMG(x,y))),  x,  y );
               }
           else if( i == cdim_y-1  &&  y+1 < img_size_y ) {
               IMG( x, y ) = (BYTE)( .6*v + .4*IMG( x, y+1 ));
               fast_show_rect( blimit((int )(IMG(x,y))),  x,  y );
               }
           else if( j == cdim_x  &&  x+1 < img_size_x ) {
               IMG( x, y ) = (BYTE)( .6*v + .4*IMG( x+1, y ));
               fast_show_rect( blimit((int )(IMG(x,y))),  x,  y );
               }
           else {
               IMG( x,  y ) = (BYTE )v;
               fast_show_rect( v,  x,  y );
               }
           }
       }
   libera( c );
   }
fclose( fp );
}


/************/
fast_show_rect( int col, int x, int y )    /* this proc. can be commented to
									speed up the compression, it shows
									the rang and domains during 
									map computation */
/************/
{
int    t, v;

t = x*zoom;  v = y*zoom;
color( col ); 
rectfi( t, v, t+zoom-1, v+zoom-1 );
}





/************/
show_map( char *mapname )
/************/
{
RECT_DATA  *c, *m, *t;
RECT   r, d;
FILE   *fp;
double o, s, rms, rp;
int    mode, i, j, n, cdim_x, cdim_y, tdim_x, rx, ry, ch, cnt=0;

fp = fopen( mapname, "rb" );
load_default_par();
crea_tab();
show_cmp_par();

while( TRUE ) {
   readmap();
   r.x2 = r.x1 + rx - 1,  r.y2 = r.y1 + ry - 1;

if( r.y1 > 230   ||  rx * ry < 64 )
   continue;
   logicop( XOR );
   show_box( 0, ROSSO, d.x1, d.y1, d.x1+mult_tab[rx]-1, d.y1+mult_tab[ry]-1 );
   show_box( 0, ROSSO, d.x1-1, d.y1-1, d.x1+mult_tab[rx]-1+1, d.y1+mult_tab[ry]-1+1 );
   show_box( 0, ROSSO, r.x1, r.y1, r.x2, r.y2 );
   show_box( 0, ROSSO, r.x1-1, r.y1-1, r.x2+1, r.y2+1 );
   logicop( REPLACE );
   
   if( getchar() == 0x1b )
	break;

   c = alloc_rect( &r ); cdim_x = X_DIM(c); cdim_y = Y_DIM(c);
   t = alloc_rect( &r ); tdim_x = X_DIM(t);

   /* Scala il dominio alle dimensioni del rango... */
   for( i=0;   i < cdim_y;   i++ )
       for( j=0;   j < cdim_x;   j++ ) {
           C(j,i) = IMG( d.x1+mult_tab[j], d.y1+mult_tab[i] ); /* dominio scalato... */
           T(j,i) = IMG( r.x1+j,           r.y1+i );           /* rango... */
           }
   m = map_mode( mode, c );        /* dominio "trasformato"... */

   n = A_DIM( m );
   for( rms=0.0, i=0;   i<n;   i++ ) {
       rp = ( (s * m->data[i] + o) - (double )t->data[i] );
       rms += (rp * rp);
       }
   rms /= n;

   logicop( XOR );
   show_box( 0, ROSSO, d.x1, d.y1, d.x1+mult_tab[rx]-1, d.y1+mult_tab[ry]-1 );
   show_box( 0, ROSSO, d.x1-1, d.y1-1, d.x1+mult_tab[rx]-1+1, d.y1+mult_tab[ry]-1+1 );
   show_box( 0, ROSSO, r.x1, r.y1, r.x2, r.y2 );
   show_box( 0, ROSSO, r.x1-1, r.y1-1, r.x2+1, r.y2+1 );
   logicop( REPLACE );
   
   libera( c );  libera( t );  libera( m );
   }
fclose( fp );
}




/************/
show_map_shape( int col, char *mapname )
/************/
{
FILE   *fp;
RECT   r, d;
double o, s;
int    mode, i, rx, ry, cnt=0;

fp = fopen( mapname, "rb" );
load_default_par();
show_cmp_par();
OPENGRAPH( "Show Grey Map Shape" );

while( TRUE ) {
   readmap();
   r.x2 = r.x1 + rx - 1,  r.y2 = r.y1 + ry - 1;
   show_box( col, NERO, r.x1, r.y1, r.x2, r.y2 );
   cnt++;
   }
MSG2VAL( "Number of maps : %d ( average area : %d )", cnt, ((img_size_x * img_size_y / cnt)+1) );

CLOSEGRAPH();
fclose( fp );
}




/*
** The  macros "subdivide()" e "writemap()" deletes from the list structure 
** an element previously considered 
*/

#define    COMPLEX_AREA        ( complexity( r ) > par_cmp.complex )
#define    BIG_AREA            ( rect_area( r ) > par_cmp.min_div_area )
#define    HUGE_AREA           ( rect_area( r ) > MAX_AREA_RANGO )
#define    TINY_AREA           ( rect_area( r ) < 32 )
#define    HUGE_ERROR          ( rms > 4 * par_cmp.max_rms )
#define    BIG_ERROR           ( rms > par_cmp.max_rms )
#define    BIG_CONTRAST        ( fabs(s) > par_cmp.max_contrast )

/************************/
/************************/
comprimi_img( char *outmap )
/************************/
/************************/
{
RECT   *r, *d, t;
double o, s, rms;
int    mode, dt=0;
FILE   *fp;


make_D_img();
Sobel_cerca_contorni();
sub_div_4( 0, NERO, 0, 0, img_size_x-1,  img_size_y-1 );
crea_tab();
show_cmp_par();
writeheader();

while( lst ) {                         /* finche' ci sono oggetti nella lista... */
   r = get_list_head( lst );           /* prendi il primo elemento della lista... */
   show_box( 0, ROSSO, r->x1, r->y1, r->x2, r->y2 );

   if( HUGE_AREA || COMPLEX_AREA ) { subdivide(); continue; }

   /****************************************** */
   d = find_D_for_R( r, &mode, &o, &s, &rms );  /* cerca un dominio "d" per "r"... */
   /****************************************** */

   if( (BIG_ERROR && BIG_AREA) || (HUGE_ERROR && !TINY_AREA) ) { 
       subdivide(); 
       }
   else {
       show_box( BIANCO, NERO, r->x1, r->y1, r->x2, r->y2 );
       /* MSG1VAL( "RMS: %.2lf", rms  ); */
       /* MSG1VAL( "Mappe create: %d ", dt++ ); */
	dt++;
       writemap();
       }
   }
MSG1VAL( "\n\n\nMappe create: %d", dt );
fclose( fp );
}




/*
** Sort in descending order...
*/
/**************************************/
int    rd_comp( const void *a, const void *b )
/**************************************/
{
if( ((RAN_DOM *)a)->rms > ((RAN_DOM *)b)->rms )
   return( -1 );
else
   return( 1 );
}



/************************/
/************************/
comprimi_img_2( char *outmap )
/************************/
/************************/
{
RECT       t, *r = &t, *d;
double     o, s, rms;
int        mode, dt=0, i, j, dx, dy;
FILE       *fp;

make_D_img();
crea_tab();
show_cmp_par();
writeheader();

/* Suddivisione iniziale... */
for( nbox=0, i=0;  i<8;  i++ )
   for( j=0;  j<8;  j++ ) {

       r->x1 = j*32, r->y1 = i*32, r->x2 = r->x1+31, r->y2 = r->y1+31;
       d = find_D_for_R( r, &mode, &o, &s, &rms );
       show_box( 0, ROSSO, r->x1, r->y1, r->x2, r->y2 );   
       rda[nbox].dominio = *d; rda[nbox].rango = *r;
       rda[nbox].o = o; rda[nbox].s = s;
       rda[nbox].mode = mode; rda[nbox].rms = rms;
       nbox++;
       }
qsort( rda, nbox, sizeof(RAN_DOM), rd_comp );

while( nbox < maxbox ) {
   t = rda[0].rango;           /* prendi l'elemento con errore massimo... */

   MSG1VAL( "Max RMS: %.2lf", rda[0].rms  );
   MSG1VAL( "Mappe create: %d ", nbox );

   if( (t.x2-t.x1+1) * (t.y2-t.y1+1) <= MIN_A ) {
       rda[0].rms = 0.0;               /* non considerare piu' le aree "tiny" */
       qsort( rda, nbox, sizeof(RAN_DOM), rd_comp );
       continue;
       }

   dx=(t.x1+t.x2)/2, dy=(t.y1+t.y2)/2;
   rda[0].rango.x1 = t.x1; rda[0].rango.x2 = dx; 
   rda[0].rango.y1 = t.y1; rda[0].rango.y2 = dy;
   d = find_D_for_R( &rda[0].rango, &mode, &o, &s, &rms );
   show_box( 0, GIALLO, rda[0].rango.x1, rda[0].rango.y1, rda[0].rango.x2, rda[0].rango.y2 );  
   rda[0].dominio = *d; rda[0].o = o; rda[0].s = s;
   rda[0].mode = mode; rda[0].rms = rms;

   rda[nbox].rango.x1 = t.x1; rda[nbox].rango.x2 = dx; 
   rda[nbox].rango.y1 = dy+1; rda[nbox].rango.y2 = t.y2;
   d = find_D_for_R( &rda[nbox].rango, &mode, &o, &s, &rms );
   show_box( 0, GIALLO, rda[nbox].rango.x1, rda[nbox].rango.y1, rda[nbox].rango.x2, rda[nbox].rango.y2 );  
   rda[nbox].dominio = *d; rda[nbox].o = o; rda[nbox].s = s;
   rda[nbox].mode = mode; rda[nbox].rms = rms;
   nbox++;

   rda[nbox].rango.x1 = dx+1; rda[nbox].rango.x2 = t.x2; 
   rda[nbox].rango.y1 = t.y1; rda[nbox].rango.y2 = dy;
   d = find_D_for_R( &rda[nbox].rango, &mode, &o, &s, &rms );
   show_box( 0, GIALLO, rda[nbox].rango.x1, rda[nbox].rango.y1, rda[nbox].rango.x2, rda[nbox].rango.y2 );  
   rda[nbox].dominio = *d; rda[nbox].o = o; rda[nbox].s = s;
   rda[nbox].mode = mode; rda[nbox].rms = rms;
   nbox++;

   rda[nbox].rango.x1 = dx+1; rda[nbox].rango.x2 = t.x2; 
   rda[nbox].rango.y1 = dy+1; rda[nbox].rango.y2 = t.y2;
   d = find_D_for_R( &rda[nbox].rango, &mode, &o, &s, &rms );
   show_box( 0, GIALLO, rda[nbox].rango.x1, rda[nbox].rango.y1, rda[nbox].rango.x2, rda[nbox].rango.y2 );  
   rda[nbox].dominio = *d; rda[nbox].o = o; rda[nbox].s = s;
   rda[nbox].mode = mode; rda[nbox].rms = rms;
   nbox++;

   qsort( rda, nbox, sizeof(RAN_DOM), rd_comp );
   }

for( i=0;  i<nbox;  i++ ) {
   fprintf( fp, "%d\t\t%d\t\t%d\t\t%d\t\t%d\t\t%d\t\t%d\t\t%d\t\t%.3f\n",
       rda[i].rango.x1,  rda[i].rango.y1,  
       rda[i].rango.x2-rda[i].rango.x1+1,  rda[i].rango.y2-rda[i].rango.y1+1,
       rda[i].dominio.x1,  rda[i].dominio.y1,  
       rda[i].mode,  (int )(rda[i].o+0.5), (float )rda[i].s );
   }
fclose( fp );
}




#define    fill_rango()    tdim_x = X_DIM(c);  tdim_y = Y_DIM(c); \
                           for( i=0;   i < tdim_y;   i++ ) \
                               for( j=0;   j < tdim_x;   j++ ) \
                                   T( j, i ) = IMG( r->x1+j,  r->y1+i );

#define    fill_dominio()      cdim_x = X_DIM(c);  cdim_y = Y_DIM(c); \
                               for( qi=0;   qi < cdim_y;   qi++ ) \
                                   for( qj=0;   qj < cdim_x;   qj++ ) \
                                       C( qj, qi ) = D_IMG( j+qj,  i+qi );

/* */
/*  For any range it searches a domain (the best) among all the */
/* possible ones, with dimensions (H,W) defined by "par_cmp.D_R_rapp". */
/* The returned parameter "mode" contains the selected transformation, that, */
/* with coeefficients "s" and "o" maps the domain into the range */
/* */
/***********/
RECT  *find_D_for_R( RECT *r, int *mode, double *bo, double *bs, double *err )
/***********/
{
static RECT    d, last_rect;
RECT_DATA      *c, *t, *ranghi[4];
double         s, o, rms, last_rms=1000009.;
int            md, i, j, n, qi, qj, dx, dy, cr, cd, 
               tdim_x, tdim_y, cdim_x, cdim_y;

   *err = 1000000.;        /* se non trova alcun dominio "buono", esce e divide... */
   dx = r->x2 - r->x1 + 1; dy = r->y2 - r->y1 + 1;
   c = alloc_rect( r ); t = alloc_rect( r ); 
   fill_rango();

   /***** time opt 1 */
   for( i=0;   i < 4;   i++ ) {      /* 8 trasf: ultime 4 utili solo per quadrati... */
       ranghi[i] = map_mode( i, t ); /* prepara ranghi "trasformati" (1 sola volta)... */
       /***** time opt 2 */
       g_b_S[i] = g_b2_S[i] = 0;
       n = A_DIM( ranghi[i] );
       for( j=0;   j < n;   j++ ) {
           cr = (int )ranghi[i]->data[j];
           g_b_S[i]  +=  cr;
           g_b2_S[i] +=  cr * cr;
           }
       }
   /***** time opt 3 */
   /* cr = complexity( r ) + 1; */

   for( i=0;   i <= D_size_y - dy;  i+=par_cmp.search_step )
       for( j=0;   j <= D_size_x - dx;  j+=par_cmp.search_step ) {
           fill_dominio();

       /*  cd = (int )(complexity( &d ) / (DRR * DRR)) + 1; */
       /*      if( (double )(max(cr, cd)) / (double )(min( cr, cd)) > 4.0 ) */
       /*          continue; */

           /* show_D(); */
           rms = find_map_error( ranghi, c, &md, &s, &o );
           /* unshow_D(); */
   
           if( rms < last_rms   &&   fabs( s ) < par_cmp.max_contrast ) {
               last_rms = rms, *bo = o, *bs = s,  *mode = md, *err = rms;
               last_rect.x1 = j*DRR, last_rect.x2 = (j+dx)*DRR,    /* .x2: not used... */
               last_rect.y1 = i*DRR, last_rect.y2 = (i+dy)*DRR;    /* .y2: not used... */
           
               /* togliere l'istruzione seguente per una  */
               /* ricerca "totale" del dominio migliore... */
               if( rms < par_cmp.max_rms )
                   goto end;
               }
           }
end:
   libera( c );    libera( t ); 
   for( i=0;   i < 4;   i++ ) {
       libera( ranghi[i] );
       }
   return( &last_rect );
}



/* */
/* Dati un dominio e un rango di dimensioni arbitrarie ( D > R ), */
/* determina, per tutte le 8 possibilita' di "mapping" di uno nell'altro, */
/* l'errore relativo a tali trasformazioni tramite "compute_s_o_rms(...) */
/* */
/****************/
double find_map_error( RECT_DATA **rango, RECT_DATA *dominio,
                       int *mode, double *rs, double *ro )
/****************/
{
int        i;
double     s, o, rms, min_rms=1000000.;

/* Calcola gli 8 valori di RMS per il dominio "d" che mappa "r"... */
for( i=0;   i < 4;   i++ ) {   /* ...8... : salta alcune trasf. inutili per rettangoli... */

   rms = compute_s_o_rms( dominio, rango[i], &s, &o, i );

   if( min_rms > rms ) {
       min_rms = rms;
       *rs = s,   *ro = o,   *mode = i;
       }
   }
return( min_rms );
}



/* */
/* "_a" rappresenta il dominio e "_b" il rango che stiamo cercando di  */
/* approssimare: "_a" e "_b" dovrebbero contenere lo stesso numero di pixel... */
/* */
/* "s"       definisce  il  CONTRASTO... */
/* "o"       "   "   la  LUMINOSITA'... */
/* "rms"     "   "   l'  ERRORE relativo... */
/* */
/************/
double     compute_s_o_rms( RECT_DATA *_a,  RECT_DATA *_b, 
                            double *rv_s,   double *rv_o,
                            int k )
/************/
{
register int   i, n;
int            a_S=0, b_S=0, a2_S=0, b2_S=0, a_S2, ab_S=0;
double         n2, s, o, rms=0.0, sn;

n = A_DIM( _a );
n2 = n;            /* modifica: ROSS (C.P.S.) pag.248 ( n2 = n, non n*n... ) */

if( compr_mode == TARGET_COMPR )
   sn = 1.0;
else
   sn = n2;

b_S = g_b_S[ k ];      /* better performance... */
b2_S = g_b2_S[ k ];    /*  "  "     "   " */

for( i=0;   i<n;   i++ ) {
   a_S  +=  a(i);
/* b_S  +=  b(i); */
   a2_S +=  (int )(a(i)) * (int )(a(i));
/* b2_S +=  (int )(b(i)) * (int )(b(i)); */
   ab_S +=  (int )(a(i)) * (int )(b(i));
   }
a_S2 = a_S * a_S;

if( n * a2_S == a_S2 ) {
   s = 0.0;
   o = b_S / n2;
   rms = (b2_S - o*b_S) / sn;
   }
else {
   s = (n2 * ab_S - a_S * b_S) / (n2 * a2_S - a_S2);
   o = (b_S - (s * a_S)) / n2;
   rms = (b2_S + s*(s*a2_S - 2*ab_S + 2*o*a_S) + o*(o*n2 - 2*b_S)) / sn;
   }

*rv_s = s,  *rv_o = o;
return( rms );
}



/****************/
RECT_DATA *map_mode( int t, RECT_DATA *c )
/****************/
{
register int   i, j;
int            mdim_x, mdim_y, cdim_x;
RECT_DATA      *m;
RECT           r;


r.x1=X1(c), r.y1=Y1(c), r.x2=X2(c), r.y2=Y2(c);
m = alloc_rect( &r );

mdim_x = X_DIM( m ), mdim_y = Y_DIM( m );
cdim_x = X_DIM( c );

switch( t ) {
   case 0:     /* nessuna trasformazione... */
       for( i=0;   i < mdim_y;   i++ )
           for( j=0;   j < mdim_x;   j++ )
               M( j, i ) = C( j, i );
       break;

   case 1:     /* inverti le X... */
       for( i=0;   i < mdim_y;   i++ )
           for( j=0;   j < mdim_x;   j++ )
               M( j, i ) = C( mdim_x-1 - j, i );
       break;

   case 2:     /* inverti le Y... */
       for( i=0;   i < mdim_y;   i++ )
           for( j=0;   j < mdim_x;   j++ )
               M( j, i ) = C( j, mdim_y-1 - i );
       break;

   case 3:     /* inverti le X e le Y... */
       for( i=0;   i < mdim_y;   i++ )
           for( j=0;   j < mdim_x;   j++ )
               M( j, i ) = C( mdim_x-1 - j, mdim_y-1 - i );
       break;

   case 4:     /* "ruota" di +90 e inverti le Y... */
       for( i=0;   i < mdim_y;   i++ )
           for( j=0;   j < mdim_x;   j++ )
               M( j, i ) = C( i,  j );
       break;

   case 5:     /* "ruota" di -90... */
       for( i=0;   i < mdim_y;   i++ )
           for( j=0;   j < mdim_x;   j++ )
               M( j, i ) = C( i,   mdim_x-1 - j );
       break;

   case 6:     /* "ruota" di +90... */
       for( i=0;   i < mdim_y;   i++ )
           for( j=0;   j < mdim_x;   j++ )
               M( j, i ) = C( mdim_y-1 - i,   j );
       break;

   case 7:     /* "ruota" di -90 e inverti le Y... */
       for( i=0;   i < mdim_y;   i++ )
           for( j=0;   j < mdim_x;   j++ )
               M( j, i ) = C( mdim_y-1 - i,   mdim_x-1 - j );
       break;
   }

return( m );
}



/************/
complexity( RECT *r )
/************/
{
register int   i, j, cnt=0;

for( i=r->y1;   i < r->y2;   i++ )
   for( j=r->x1;   j < r->x2;   j++ )
       if( IMG2( j, i ) ) cnt++;

return( cnt );
}



/************/
RECT_DATA  *alloc_rect( RECT *r )
/************/
{
RECT_DATA  *c;

c = (RECT_DATA *)malloc( sizeof(RECT_DATA) );
c->vertex = *r;
c->data = (BYTE *)malloc( A_DIM( c ) );

return( c );
}



/**************/
get_dim_fm_map( char *mapname )
/**************/
{
FILE   *fp;

if( (fp = fopen( mapname, "rb" )) == NULL ) {
   MSG1VAL( "Cannot open %s", mapname );
   return( 0 );
   }
fscanf( fp, "%d %d", &img_size_x, &img_size_y );
fclose( fp );

return( 1 );
}



/************** */
show_cmp_par()
/************** */
{
MSG2VAL( "\n\nImg dimension: dx=%d, dy=%d", img_size_x, img_size_y );
MSG1VAL( "RMS limit: %.1lf", par_cmp.max_rms );
MSG1VAL( "dim(D) / dim(R): %.1lf", par_cmp.D_R_rapp );
MSG1VAL( "Max contrast: %.1lf", par_cmp.max_contrast );
MSG1VAL( "Min range area: %d", par_cmp.min_div_area );
MSG1VAL( "Threshold for edges : %d", par_cmp.soglia );
MSG1VAL( "Search step: %d", par_cmp.search_step );
MSG1VAL( "Box complexity break: %d\n\n", par_cmp.complex );
}



/********** */
crea_tab()
/********** */
{
int    i;

for( i=0;  i<1280;  i++ )
   mult_tab[i] = (int )(i * par_cmp.D_R_rapp);
}




/********** */
make_D_img()
/********** */
{
register int   i, j;

D_size_x = img_size_x / par_cmp.D_R_rapp,
D_size_y = img_size_y / par_cmp.D_R_rapp;
D_scaled = (BYTE *)malloc( D_size_x * D_size_y );

for( i=0;   i < D_size_y;  i++ )
   for( j=0;   j < D_size_x;  j++ ) {
       D_IMG( j, i ) = (BYTE )(((int )IMG( mult_tab[j], mult_tab[i] ) +
                       (int )IMG( mult_tab[j]+1, mult_tab[i] ) +
                       (int )IMG( mult_tab[j], mult_tab[i]+1 ) +
                       (int )IMG( mult_tab[j]+1, mult_tab[i]+1 ) ) / 4 );
       /* show_rect( blimit(D_IMG( j, i )), img_size_x+j+30, i ); */
       }
}



/*************/
get_map_name( char *mapf )
/*************/
{
char   s[128], rgb[8];

   ask_str( "\nMap Name [no ext]: ", mapf );
   strcat( mapf, ".map" );

   if( ! exist( mapf ) ) {
       printf("\n\nCannot open %s\n", mapf );
       beep(); return( 0 );
       }
   else
       return( 1 );
}



/*************/
get_img_name( char *imgf )
/*************/
{
char   s[128], rgb[8];

   ask_str( "\nImage Name [no ext]: ", imgf );
   strcat( imgf, ".sgi" );

   if( ! exist( imgf ) ) {
       printf("\n\nCannot open %s\n", imgf );
       beep();  return( 0 );
       }
   else
       return( 1 );
}



write_map_record(  FILE *fp, int rsx, int rsy, int rdx, int rdy, 
                   int dsx, int dsy, int modo, int brigh, int contr )
{
DATA_OUT   a;

a.rsx = rsx; a.rsy = rsy;
a.rdx = rdx; a.rdy = rdy;
a.dsx = dsx; a.dsy = dsy;
a.modo = modo;
a.brigh = max( -2047, min( 2047, brigh ));     /* 12 bit */
a.contr = max( -2047, min( 2047, contr ));

fwrite( &a, sizeof(DATA_OUT), 1, fp );
}


read_map_record(   FILE *fp, int *rsx, int *rsy, int *rdx, int *rdy, 
                   int *dsx, int *dsy, int *modo, double *brigh, double *contr )
{
DATA_OUT   a;
int    i;

i = fread( &a, sizeof(DATA_OUT), 1, fp );

*rsx = a.rsx; *rsy = a.rsy;
*rdx = a.rdx; *rdy = a.rdy;
*dsx = a.dsx; *dsy = a.dsy;
*modo = a.modo;
*brigh = (double )a.brigh;
*contr = (double )a.contr / 100.;

return( i );
}

